import os
from pathlib import Path
from typing import Optional

from prospector.formatters.base import Formatter
from prospector.message import Location, Message


class SummaryFormatter(Formatter):
    """
    This abstract formatter is used to output a summary of the prospector run.
    """

    summary_labels = (
        ("started", "Started", None),
        ("completed", "Finished", None),
        ("time_taken", "Time Taken", lambda x: f"{x} seconds"),
        ("formatter", "Formatter", None),
        ("profiles", "Profiles", None),
        ("strictness", "Strictness", None),
        ("libraries", "Libraries Used", ", ".join),
        ("tools", "Tools Run", ", ".join),
        ("adaptors", "Adaptors", ", ".join),
        ("message_count", "Messages Found", None),
        ("external_config", "External Config", None),
    )

    def render_summary(self) -> str:
        output = [
            "Check Information",
            "=================",
        ]

        label_width = max(len(label[1]) for label in self.summary_labels)

        for key, label, formatter in self.summary_labels:
            if key in self.summary:
                value = self.summary[key]
                if formatter is not None:
                    value = formatter(value)
                output.append(f" {label.rjust(label_width)}: {value}")

        return "\n".join(output)

    def render_profile(self) -> str:
        output = ["Profile", "=======", "", self.profile.as_yaml().strip()]

        return "\n".join(output)

    def get_ci_annotation(self, message: Message) -> Optional[str]:
        intro = (
            f"({message.source})"
            if message.code is None
            else f"{message.code}({message.source})"
            if message.location.function is None
            else f"[{message.code}({message.source}), {message.location.function}]"
        )
        ci_prefix = self._get_ci_prefix(message.location, intro)
        if ci_prefix:
            github_message = f"{ci_prefix}{intro}%0A{message.message.strip()}"
            if message.doc_url:
                github_message += f"%0ASee: {message.doc_url}"
            return github_message
        return None

    def _get_ci_prefix(self, location: Location, title: str) -> Optional[str]:
        if location.path is None:
            return None
        if os.environ.get("GITHUB_ACTIONS") == "true":
            path = location.path
            if "PROSPECTOR_FILE_PREFIX" in os.environ:
                path = Path(os.environ["PROSPECTOR_FILE_PREFIX"]) / path
            # pylint: disable-next=line-too-long
            # See: https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/workflow-commands-for-github-actions#setting-an-error-message
            parameters = {
                "file": path,
                "line": location.line,
                "title": title,
            }
            if location.line_end:
                parameters["endLine"] = location.line_end
            if location.character:
                parameters["col"] = location.character
            if location.character_end:
                parameters["endColumn"] = location.character_end

            parameters_str = ",".join(f"{key}={value}" for key, value in parameters.items())
            return f"::error {parameters_str}::"

        return None



from typing import TYPE_CHECKING, Any, Optional

from bandit.cli.main import _get_profile, _init_extensions
from bandit.core import docs_utils
from bandit.core.config import BanditConfig
from bandit.core.constants import RANKING
from bandit.core.manager import BanditManager

from prospector.finder import FileFinder
from prospector.message import Location, Message
from prospector.tools.base import ToolBase

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig


class BanditTool(ToolBase):
    manager: Optional[BanditManager] = None
    profile: Optional[str] = None
    config_file: Optional[str] = None
    agg_type = "file"
    severity = 0
    confidence = 0

    def configure(self, prospector_config: "ProspectorConfig", _: Any) -> None:
        options = prospector_config.tool_options("bandit")

        if "profile" in options:
            self.profile = options.pop("profile")

        if "config" in options:
            self.config_file = options.pop("config")

        if "severity" in options:
            self.severity = options.pop("severity")
            if not 0 <= self.severity <= 2:
                raise ValueError(f"severity {self.severity!r} must be between 0 and 2")

        if "confidence" in options:
            self.confidence = options.pop("confidence")
            if not 0 <= self.confidence <= 2:
                raise ValueError(f"confidence {self.confidence!r} must be between 0 and 2")

        b_conf = BanditConfig(config_file=self.config_file)
        disabled_messages = prospector_config.get_disabled_messages("bandit")
        if disabled_messages:
            b_conf.config.setdefault("skips", []).extend(disabled_messages)
        if options:
            b_conf.config.update(options)
            b_conf.validate(path="<prospector config>")
        profile = _get_profile(b_conf, self.profile, self.config_file)
        extension_mgr = _init_extensions()
        extension_mgr.validate_profile(profile)

        self.manager = BanditManager(b_conf, None, profile=profile)

    def run(self, found_files: FileFinder) -> list[Message]:
        assert self.manager is not None

        self.manager.files_list = sorted(found_files.files)
        self.manager.exclude_files = []

        if not self.manager.b_ts.tests:
            raise ValueError("No test will run for bandit")

        self.manager.run_tests()
        results = self.manager.get_issue_list(sev_level=RANKING[self.severity], conf_level=RANKING[self.confidence])
        messages = []
        for result in results:
            loc = Location(
                result.fname,
                None,
                "",
                result.lineno,
                result.col_offset,
                line_end=result.linerange[-1] if result.linerange else result.lineno,
                character_end=result.end_col_offset,
            )
            msg = Message("bandit", result.test_id, loc, result.text, doc_url=docs_utils.get_url(result.test_id))
            messages.append(msg)
        return messages



import json
import subprocess  # nosec
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ruff.__main__ import find_ruff_bin

from prospector.finder import FileFinder
from prospector.message import Location, Message
from prospector.tools.base import PEP8_IGNORE_LINE_CODE, ToolBase

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig


class RuffTool(ToolBase):
    def configure(self, prospector_config: "ProspectorConfig", _: Any) -> None:
        self.ruff_bin = find_ruff_bin()
        self.ruff_args = ["check", "--output-format=json"]

        enabled = prospector_config.get_enabled_messages("ruff")
        if enabled:
            enabled_arg_value = ",".join(enabled)
            self.ruff_args.append(f"--select={enabled_arg_value}")
        disabled = prospector_config.get_disabled_messages("ruff")
        if disabled:
            disabled_arg_value = ",".join(disabled)
            self.ruff_args.append(f"--ignore={disabled_arg_value}")

        options = prospector_config.tool_options("ruff")
        for key, value in options.items():
            if value is True:
                self.ruff_args.append(f"--{key}")
            elif value is False:
                pass
            elif isinstance(value, list):
                arg_value = ",".join(value)
                self.ruff_args.append(f"--{key}={arg_value}")
            # dict is like array but with a dict with true/false value to be able to merge profiles
            elif isinstance(value, dict):
                arg_value = ",".join(k for k, v in value.items() if v)
                self.ruff_args.append(f"--{key}={arg_value}")
            else:
                self.ruff_args.append(f"--{key}={value}")

    def run(self, found_files: FileFinder) -> list[Message]:
        messages = []
        completed_process = subprocess.run(  # noqa
            [self.ruff_bin, *self.ruff_args, *found_files.python_modules], capture_output=True
        )
        if not completed_process.stdout:
            messages.append(
                Message(
                    "ruff",
                    "",
                    Location(None, None, None, None, None),
                    completed_process.stderr.decode(),
                )
            )
            return messages
        for message in json.loads(completed_process.stdout):
            sub_message = {}
            if message.get("fix") and message["fix"].get("applicability"):
                sub_message["Fix applicability"] = message["fix"]["applicability"]
            message_str = message.get("message", "")
            if sub_message:
                message_str += f" [{', '.join(f'{k}: {v}' for k, v in sub_message.items())}]"

            if message.get("filename") is None or found_files.is_excluded(Path(message.get("filename"))):
                continue
            messages.append(
                Message(
                    "ruff",
                    message.get("code") or "unknown",
                    Location(
                        message.get("filename") or "unknown",
                        None,
                        None,
                        line=message.get("location", {}).get("row"),
                        character=message.get("location", {}).get("column"),
                        line_end=message.get("end_location", {}).get("row"),
                        character_end=message.get("end_location", {}).get("column"),
                    ),
                    message_str,
                    doc_url=message.get("url"),
                    is_fixable=bool((message.get("fix") or {}).get("applicability") in ("safe", "unsafe")),
                )
            )
        return messages

    def get_ignored_codes(self, line: str) -> list[tuple[str, int]]:
        match = PEP8_IGNORE_LINE_CODE.search(line)
        if match:
            return [(e.strip(), 0) for e in match.group(1).split(",")]
        return []



import json
import subprocess  # nosec
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional

import pyright

from prospector.finder import FileFinder
from prospector.message import Location, Message
from prospector.tools import ToolBase
from prospector.tools.exceptions import BadToolConfig

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig


__all__ = ("PyrightTool",)


VALID_OPTIONS = [
    "level",
    "project",
    "pythonplatform",
    "pythonversion",
    "skipunannotated",
    "typeshed-path",
    "venv-path",
]


def format_messages(json_encoded: str) -> list[Message]:
    json_decoded = json.loads(json_encoded)
    diagnostics = json_decoded.get("generalDiagnostics", [])
    messages = []
    for diag in diagnostics:
        start_range = diag.get("range", {}).get("start", {})
        location = Location(
            path=diag["file"],
            module=None,
            function=None,
            line=start_range.get("line", -1),
            character=start_range.get("character", -1),
        )
        messages.append(
            Message(source="pyright", code=diag.get("rule", ""), location=location, message=diag.get("message", ""))
        )

    return messages


class PyrightTool(ToolBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.checker = pyright
        self.options = ["--outputjson"]

    def configure(  # pylint: disable=useless-return
        self, prospector_config: "ProspectorConfig", _: Any
    ) -> Optional[tuple[str, Optional[Iterable[Message]]]]:
        options = prospector_config.tool_options("pyright")

        for option_key in options:
            if option_key not in VALID_OPTIONS:
                url = "https://github.com/PyCQA/prospector/blob/master/prospector/tools/pyright/__init__.py"
                raise BadToolConfig(
                    "pyright", f"Option {option_key} is not valid. See the list of possible options: {url}"
                )

        level = options.get("level", None)
        project = options.get("project", None)
        pythonplatform = options.get("pythonplatform", None)
        pythonversion = options.get("pythonversion", None)
        skipunannotated = options.get("skipunannotated", False)
        typeshed_path = options.get("typeshed-path", None)
        venv_path = options.get("venv-path", None)

        if level:
            self.options.extend(["--level", level])
        if project:
            self.options.extend(["--project", project])
        if pythonplatform:
            self.options.extend(["--pythonplatform", pythonplatform])
        if pythonversion:
            self.options.extend(["--pythonversion", pythonversion])
        if skipunannotated:
            self.options.append("--skipunannotated")
        if typeshed_path:
            self.options.extend(["--typeshed-path", typeshed_path])
        if venv_path:
            self.options.extend(["--venv-path", venv_path])

        return None

    def run(self, found_files: FileFinder) -> list[Message]:
        paths = [str(path) for path in found_files.python_modules]
        paths.extend(self.options)
        result = self.checker.run(*paths, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        return format_messages(result.stdout)



from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional, Union

from packaging import version as packaging_version
from pylint import version as pylint_version
from pylint.config.config_initialization import _config_initialization
from pylint.lint import PyLinter

from prospector.finder import FileFinder


class UnrecognizedOptions(Exception):
    """Raised when an unrecognized option is found in the Pylint configuration."""


class ProspectorLinter(PyLinter):
    def __init__(self, found_files: FileFinder, *args: Any, **kwargs: Any) -> None:
        self._files = found_files
        # set up the standard PyLint linter
        PyLinter.__init__(self, *args, **kwargs)

    # Largely inspired by https://github.com/pylint-dev/pylint/blob/main/pylint/config/config_initialization.py#L26
    def config_from_file(self, config_file: Optional[Union[str, Path]] = None) -> bool:
        """Initialize the configuration from a file."""
        _config_initialization(self, [], config_file=config_file)
        return True

    def _expand_files(self, files_or_modules: list[str]) -> Union[Iterable[Any], dict[str, Any]]:
        expanded = super()._expand_files(files_or_modules)
        filtered: dict[str, Any] = {}
        # PyLinter._expand_files returns dict since 2.15.7.
        if packaging_version.parse(pylint_version) > packaging_version.parse("2.15.6"):
            for module, expanded_module in expanded.items():
                if not self._files.is_excluded(Path(module)):
                    filtered[module] = expanded_module
            return filtered
        else:
            for module in expanded:
                # need to de-duplicate, as pylint also walks directories given to it, so it will find
                # files that prospector has already provided and end up checking it more than once
                if not self._files.is_excluded(Path(module["path"])):
                    # if the key exists, just overwrite it with the same value, so we don't need an extra if statement
                    filtered[module["path"]] = module
            return filtered.values()



import mimetypes
import os
import re
from pathlib import Path

from prospector import identify

_PYTHON_COMMAND_RE = re.compile(r"^python[0-9]?$")


def is_python_package(path: Path) -> bool:
    return path.is_dir() and (path / "__init__.py").exists()


def is_python_module(path: Path) -> bool:
    mimetype, encoding = mimetypes.guess_type(path)
    del encoding
    if mimetype == "text/x-python":
        return True

    executor = identify.parse_shebang_from_file(path)
    if executor is not None and len(executor) > 0:
        return _PYTHON_COMMAND_RE.match(Path(executor[0]).name) is not None

    return False


def is_virtualenv(path: Path) -> bool:
    clues = ("Scripts", "lib", "include") if os.name == "nt" else ("bin", "lib", "include")

    try:
        # just get the name, iterdir returns absolute paths by default
        dircontents = [obj.name for obj in path.iterdir()]
    except (OSError, TypeError):
        # listdir failed, probably due to path length issues in windows
        return False

    if not all(clue in dircontents for clue in clues):
        # we don't have the 3 directories which would imply
        # this is a virtualenvironment
        return False

    if not all((path / clue).is_dir() for clue in clues):
        # some of them are not actually directories
        return False

    # if we do have all three directories, make sure that it's not
    # just a coincidence by doing some heuristics on the rest of
    # the directory
    # if there are more than 7 things it's probably not a virtualenvironment
    return len(dircontents) <= 7



import os
import re

from prospector.formatters.base_summary import SummaryFormatter


class PylintFormatter(SummaryFormatter):
    """
    This formatter outputs messages in the same way as pylint -f parseable , which is used by several
    tools to parse pylint output. This formatter is therefore a compatibility shim between tools built
    on top of pylint and prospector itself.
    """

    def render_messages(self) -> list[str]:
        cur_loc = None
        output = []
        for message in sorted(self.messages):
            if cur_loc != message.location.path:
                cur_loc = message.location.path
                module_name = str(self._make_path(message.location)).replace(os.path.sep, ".")
                module_name = re.sub(r"(\.__init__)?\.py$", "", module_name)

                header = f"************* Module {module_name}"
                output.append(header)

            #   ={path}:{line}:{character}: [{msg_id}({symbol}), {obj}] {msg}
            # prospector/configuration.py:65:1: [missing-docstring(missing-docstring), build_default_sources] \
            # Missing function docstring

            template_location = (
                ""
                if message.location.path is None
                else "%(path)s"
                if message.location.line is None
                else "%(path)s:%(line)s"
                if message.location.character is None
                else "%(path)s:%(line)s:%(character)s"
            )
            template_code = (
                "(%(source)s)"
                if message.code is None
                else "%(code)s(%(source)s)"
                if message.location.function is None
                else "[%(code)s(%(source)s), %(function)s]"
            )
            template = (
                f"{template_location}: {template_code}: %(message)s"
                if template_location
                else f"{template_code}: %(message)s"
            )

            message_str = message.message.strip()
            if message.doc_url:
                message_str += f" (See: {message.doc_url})"
            template_args = {
                "path": self._make_path(message.location),
                "line": message.location.line,
                "character": message.location.character,
                "source": message.source,
                "code": message.code,
                "function": message.location.function,
                "message": message.message.strip(),
            }
            output.append(template % template_args)
            if message.doc_url:
                template_args["message"] = ""
                indent = len(template % template_args)
                output.append(f"{' ' * indent}See: {message.doc_url}")
            ci_annotation = self.get_ci_annotation(message)
            if ci_annotation:
                output.append(ci_annotation)
        return output

    def render(self, summary: bool = True, messages: bool = True, profile: bool = False) -> str:
        output: list[str] = []
        if messages:
            output.extend(self.render_messages())
        if profile:
            output.append("")
            output.append(self.render_profile())
        if summary:
            output.append("")
            output.append(self.render_summary())

        return "\n".join(output)



import os
import re
import warnings
from pathlib import Path
from typing import Union

from requirements_detector import find_requirements
from requirements_detector.detect import RequirementsNotFound

from prospector import encoding
from prospector.exceptions import PermissionMissing
from prospector.pathutils import is_virtualenv

POSSIBLE_LIBRARIES = ("django", "celery")


# see http://docs.python.org/2/reference/lexical_analysis.html#identifiers
_FROM_IMPORT_REGEX = re.compile(r"^\s*from ([\._a-zA-Z0-9]+) import .*$")
_IMPORT_REGEX = re.compile(r"^\s*import ([\._a-zA-Z0-9]+)$")
_IMPORT_MULTIPLE_REGEX = re.compile(r"^\s*import ([\._a-zA-Z0-9]+(, ){1})+")


def find_from_imports(file_contents: str) -> set[str]:
    names = set()
    for line in file_contents.split("\n"):
        match = _IMPORT_MULTIPLE_REGEX.match(line)
        if match:
            import_names = []
            first = match.group(1)
            import_names.append(first[:-2])
            for name in line.split(first)[1].split(","):
                import_names.append(name.strip())
        else:
            match = _IMPORT_REGEX.match(line) or _FROM_IMPORT_REGEX.match(line)
            if match is None:
                continue
            import_names = match.group(1).split(".")

        for import_name in import_names:
            if import_name in POSSIBLE_LIBRARIES:
                names.add(import_name)

    return names


def find_from_path(path: Path) -> set[str]:
    names = set()

    try:
        for item in path.iterdir():
            if item.is_dir():
                if is_virtualenv(item):
                    continue
                names |= find_from_path(item)
            elif not item.is_symlink() and item.suffix == ".py":
                try:
                    contents = encoding.read_py_file(item)
                    names |= find_from_imports(contents)
                except encoding.CouldNotHandleEncoding as err:
                    # TODO: this output will break output formats such as JSON
                    warnings.warn(f"{err.path}: {err.__cause__}", ImportWarning, stacklevel=0)

            if len(names) == len(POSSIBLE_LIBRARIES):
                # don't continue on recursing, there's no point!
                break
    except PermissionError as err:
        raise PermissionMissing(path) from err

    return names


def find_from_requirements(path: Union[str, Path]) -> set[str]:
    reqs = find_requirements(path)
    names: set[str] = set()
    for requirement in reqs:
        if requirement.name is not None and requirement.name.lower() in POSSIBLE_LIBRARIES:
            names.add(requirement.name.lower())
    return names


def autodetect_libraries(path: Union[str, Path]) -> set[str]:
    if os.path.isfile(path):
        path = os.path.dirname(path)
        if path == "":
            path = "."

    libraries: set[str] = set()

    try:
        libraries = find_from_requirements(path)
    except RequirementsNotFound:
        pass

    if len(libraries) < len(POSSIBLE_LIBRARIES):
        libraries = find_from_path(Path(path))

    return libraries



from prospector.formatters.text import TextFormatter
from prospector.message import Message

__all__ = ("EmacsFormatter",)


class EmacsFormatter(TextFormatter):
    def render_message(self, message: Message) -> str:
        output = [
            f"{self._make_path(message.location)}:{message.location.line}:{(message.location.character or 0) + 1}:",
            "    L{}:{} {}: {} - {}".format(
                message.location.line or "-",
                message.location.character if message.location.line else "-",
                message.location.function,
                message.source,
                message.code,
            ),
            f"    {message.message}",
        ]

        return "\n".join(output)



import re
from abc import ABC, abstractmethod
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from prospector.finder import FileFinder
from prospector.message import Message

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig

PEP8_IGNORE_LINE_CODE = re.compile(r"#\s*noqa:([^#]*[^# ])(\s*#.*)?$", re.IGNORECASE)


class ToolBase(ABC):
    @abstractmethod
    def configure(
        self, prospector_config: "ProspectorConfig", found_files: FileFinder
    ) -> Optional[tuple[Optional[Union[str, Path]], Optional[Iterable[Message]]]]:
        """
        Tools have their own way of being configured from configuration files
        on the current path - for example, a .pep8rc file. Prospector will use
        its own configuration settings unless this method discovers some
        tool-specific configuration that should be used instead.

        :return: A tuple: the first element is a string indicating how or where
                 this tool was configured from. For example, this can be a path
                 to the .pylintrc file used, if used. None means that prospector
                 defaults were used. The second element should be an iterable of
                 Message objects representing any issues which were found when
                 trying to load configuration - for example, bad values in a
                 .pylintrc file. It is also possible to simply return None if
                 neither value is useful.
        """
        raise NotImplementedError

    @abstractmethod
    def run(self, found_files: FileFinder) -> list[Message]:
        """
        Actually run the tool and collect the various messages emitted by the tool.
        It is expected that this will convert whatever output of the tool into the
        standard prospector Message and Location objects.
        """
        raise NotImplementedError

    def get_ignored_codes(self, line: str) -> list[tuple[str, int]]:
        """
        Return a list of error codes and line offset that the tool will ignore from a line of code.
        """
        del line  # unused
        return []



from . import emacs, gitlab, grouped, json, pylint, pylint_parseable, text, vscode, xunit, yaml
from .base import Formatter

__all__ = ("FORMATTERS", "Formatter")


FORMATTERS: dict[str, type[Formatter]] = {
    "json": json.JsonFormatter,
    "text": text.TextFormatter,
    "gitlab": gitlab.GitlabFormatter,
    "grouped": grouped.GroupedFormatter,
    "emacs": emacs.EmacsFormatter,
    "yaml": yaml.YamlFormatter,
    "pylint": pylint.PylintFormatter,
    "pylint_parseable": pylint_parseable.PylintParseableFormatter,
    "xunit": xunit.XunitFormatter,
    "vscode": vscode.VSCodeFormatter,
}



from pathlib import Path


def is_relative_to(relpath: Path, otherpath: Path) -> bool:
    # is_relative_to was only added to Path in Python 3.9
    if hasattr(relpath, "is_relative_to"):
        return relpath.is_relative_to(otherpath)

    try:
        relpath.relative_to(otherpath)
    except ValueError:
        return False
    else:
        return True



import hashlib
import json
from typing import Any

from prospector.formatters.base import Formatter

__all__ = ("GitlabFormatter",)


class GitlabFormatter(Formatter):
    """
    This formatter outputs messages in the GitLab Code Quality report format.
    https://docs.gitlab.com/ci/testing/code_quality/#code-quality-report-format
    """

    def render(self, summary: bool = True, messages: bool = True, profile: bool = False) -> str:
        output: list[dict[str, Any]] = []
        fingerprints = set()

        if messages:
            for message in sorted(self.messages):
                # Make sure that we do not get a fingerprint that is already in use
                # by adding in the previously generated one.
                message_hash = ":".join([str(message.location.path), str(message.location.line), message.code])
                sha256_hash = hashlib.sha256(message_hash.encode())
                MAX_ITERATIONS = 1000
                iteration_count = 0
                while sha256_hash.hexdigest() in fingerprints:
                    # In cases of hash collisions, new hashes will be generated.
                    sha256_hash.update(sha256_hash.hexdigest().encode())
                    iteration_count += 1
                    if iteration_count > MAX_ITERATIONS:
                        raise RuntimeError("Maximum iteration limit reached while resolving hash collisions.")

                fingerprint = sha256_hash.hexdigest()
                fingerprints.add(fingerprint)

                output.append(
                    {
                        "type": "issue",
                        "check_name": message.code,
                        "description": f"{message.source}[{message.code}]: {message.message.strip()}",
                        "severity": "major",
                        "location": {
                            "path": str(self._make_path(message.location)),
                            "lines": {"begin": message.location.line, "end": message.location.line_end},
                        },
                        "fingerprint": fingerprint,
                    }
                )

        return json.dumps(output, indent=2)



from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Callable, Optional

from prospector.exceptions import PermissionMissing
from prospector.pathutils import is_python_module, is_python_package, is_virtualenv

_SKIP_DIRECTORIES = (".git", ".tox", ".mypy_cache", ".pytest_cache", ".venv", "__pycache__", "node_modules")


class FileFinder:
    """
    This class is responsible for taking a combination of command-line arguments
    and configuration loaded from a profile to discover all files and modules which
    should be inspected.

    Individual tools can be told to ignore certain files, so the job of this class
    is basically to know which files to pass to which tools to be inspected.
    """

    def __init__(self, *provided_paths: Path, exclusion_filters: Optional[Iterable[Callable[[Path], bool]]] = None):
        """
        :param provided_paths:
            A list of Path objects to search for files and modules - can be either directories or files
        :param exclusion_filters:
            An optional list of filters. All paths will be checked against this list - if any return True,
            the path is excluded.
        """
        self._provided_files = []
        self._provided_dirs = []
        self._exclusion_filters = [
            # we always want to ignore some things
            lambda _path: _path.is_dir() and _path.name in _SKIP_DIRECTORIES,
            is_virtualenv,
        ] + list(exclusion_filters or [])

        for path in provided_paths:
            if not path.exists():
                raise FileNotFoundError(path)
            # ensure all paths from now one are absolute paths; they can be converted
            # to relative paths for output purposes later
            path = path.absolute()
            if path.is_file():
                self._provided_files.append(path)
            if path.is_dir():
                self._provided_dirs.append(path)

    def make_syspath(self) -> list[Path]:
        paths = set()
        for path in self._provided_dirs:
            paths.add(path)
        for module in self.python_modules:
            paths.add(module.parent)
        return sorted(paths)

    def is_excluded(self, path: Path) -> bool:
        return any(filt(path) for filt in self._exclusion_filters)

    def _filter(self, paths: Iterable[Path]) -> set[Path]:
        return {path for path in paths if not self.is_excluded(path)}

    def _walk(self, directory: Path) -> Iterator[Path]:
        if not self.is_excluded(directory):
            yield directory

        for path in directory.iterdir():
            if self.is_excluded(path):
                continue
            if path.is_dir():
                yield from self._walk(path)
            else:
                yield path

    @property
    def files(self) -> set[Path]:
        """
        List every individual file found from the given configuration.

        This method is useful for tools which require an explicit list of files to check.
        """
        files = set()

        for directory in self.directories:
            for path in self._walk(directory):
                if path.is_file():
                    files.add(path)

        files = self._filter(files)

        for path in self._provided_files:
            files.add(path)

        return files

    @property
    def python_packages(self) -> list[Path]:
        """
        Lists every directory found in the given configuration which is a python module (that is,
        contains an `__init__.py` file).

        This method is useful for passing to tools which will do their own discovery of python files.
        """
        return [d for d in self.directories if is_python_package(d)]

    @property
    def python_modules(self) -> list[Path]:
        """
        Lists every directory found in the given configuration which is a python module (that is,
        contains an `__init__.py` file).

        This method is useful for passing to tools which will do their own discovery of python files.
        """
        return [f for f in self.files if is_python_module(f)]

    @property
    def directories(self) -> set[Path]:
        """
        Lists every directory found from the given configuration, regardless of its contents.

        This method is useful for passing to tools which will do their own discovery of python files.
        """
        dirs = set()
        for directory in self._provided_dirs:
            dirs.add(directory)
            try:
                for obj in self._walk(directory):
                    if obj.is_dir():
                        dirs.add(obj)
            except PermissionError as err:
                raise PermissionMissing(obj) from err

        return self._filter(dirs)



import os
import re

from prospector.formatters.base_summary import SummaryFormatter


class VSCodeFormatter(SummaryFormatter):
    """
    This formatter outputs messages in the same way as vscode prospector linter expects.
    """

    def render_messages(self) -> list[str]:
        cur_loc = None
        output = []

        for message in sorted(self.messages):
            if cur_loc != message.location.path:
                cur_loc = message.location.path
                module_name = str(self._make_path(message.location)).replace(os.path.sep, ".")
                module_name = re.sub(r"(\.__init__)?\.py$", "", module_name)

                header = f"************* Module {module_name}"
                output.append(header)

            template = "%(line)s,%(character)s,%(code)s,%(code)s:%(source)s %(message)s"
            output.append(
                template
                % {
                    "line": message.location.line,
                    "character": message.location.character,
                    "source": message.source,
                    "code": message.code,
                    "message": message.message.strip(),
                }
            )
        return output

    def render(self, summary: bool = True, messages: bool = True, profile: bool = False) -> str:
        output: list[str] = []
        if messages:
            output.extend(self.render_messages())
        if profile:
            output.append("")
            output.append(self.render_profile())
        if summary:
            output.append("")
            output.append(self.render_summary())

        return "\n".join(output)



from pathlib import Path
from typing import Optional, Union


class Location:
    _path: Optional[Path]

    def __init__(
        self,
        path: Optional[Union[Path, str]],
        module: Optional[str],
        function: Optional[str],
        line: Optional[int],
        character: Optional[int],
        line_end: Optional[int] = None,
        character_end: Optional[int] = None,
    ):
        if isinstance(path, Path):
            self._path = path.absolute()
        elif isinstance(path, str):
            self._path = Path(path).absolute()
        elif path is None:
            self._path = None
        else:
            raise ValueError
        self.module = module or None
        self.function = function or None
        self.line = None if line == -1 else line
        self.character = None if character == -1 else character
        self.line_end = line_end
        self.character_end = character_end

    @property
    def path(self) -> Optional[Path]:
        return self._path

    def absolute_path(self) -> Optional[Path]:
        return self._path

    def relative_path(self, root: Optional[Path]) -> Optional[Path]:
        if self._path is None:
            return None
        if root is None:
            return self._path
        return self._path.relative_to(root) if self._path.is_relative_to(root) else self._path

    def __repr__(self) -> str:
        return f"{self._path}:L{self.line}:{self.character}"

    def __hash__(self) -> int:
        return hash((self._path, self.line, self.character))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Location):
            return False
        return self._path == other._path and self.line == other.line and self.character == other.character

    def __lt__(self, other: "Location") -> bool:
        if not isinstance(other, Location):
            raise ValueError

        if self._path is None and other._path is None:
            return False
        if self._path is None:
            return True
        if other._path is None:
            return False
        if self._path == other._path:
            if self.line == other.line:
                return (self.character or -1) < (other.character or -1)
            return (self.line or -1) < (other.line or -1)  # line can be None if it's a file-global warning
        return self._path < other._path


class Message:
    def __init__(
        self,
        source: str,
        code: str,
        location: Location,
        message: str,
        doc_url: Optional[str] = None,
        is_fixable: bool = False,
    ):
        self.source = source
        self.code = code
        self.location = location
        self.message = message
        self.doc_url = doc_url
        self.is_fixable = is_fixable

    def __repr__(self) -> str:
        return f"{self.source}-{self.code}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Message):
            return False
        if self.location == other.location:
            return self.code == other.code
        return False

    def __lt__(self, other: "Message") -> bool:
        if self.location == other.location:
            return self.code < other.code
        return self.location < other.location


def make_tool_error_message(
    filepath: Union[Path, str],
    source: str,
    code: str,
    message: str,
    line: Optional[int] = None,
    character: Optional[int] = None,
    module: Optional[str] = None,
    function: Optional[str] = None,
) -> Message:
    location = Location(path=filepath, module=module, function=function, line=line, character=character)
    return Message(source=source, code=code, location=location, message=message)



from abc import ABC, abstractmethod

from prospector.profiles.profile import ProspectorProfile

__all__ = ("Formatter",)

from pathlib import Path
from typing import Any, Optional

from prospector.message import Location, Message


class Formatter(ABC):
    def __init__(
        self,
        summary: dict[str, Any],
        messages: list[Message],
        profile: ProspectorProfile,
        paths_relative_to: Optional[Path] = None,
    ) -> None:
        self.summary = summary
        self.messages = messages
        self.profile = profile
        self.paths_relative_to = paths_relative_to

    @abstractmethod
    def render(self, summary: bool = True, messages: bool = True, profile: bool = False) -> str:
        raise NotImplementedError

    def _make_path(self, location: Location) -> Path:
        path_ = location.relative_path(self.paths_relative_to)
        return Path() if path_ is None else path_

    def _message_to_dict(self, message: Message) -> dict[str, Any]:
        loc = {
            "path": str(self._make_path(message.location)),
            "module": message.location.module,
            "function": message.location.function,
            "line": message.location.line,
            "character": message.location.character,
        }
        if message.location.line_end is not None and message.location.line_end != -1:
            loc["lineEnd"] = message.location.line_end
        if message.location.character_end is not None and message.location.character_end != -1:
            loc["characterEnd"] = message.location.character_end
        result = {
            "source": message.source,
            "code": message.code,
            "location": loc,
            "message": message.message,
            "isFixable": message.is_fixable,
        }
        if message.doc_url:
            result["docUrl"] = message.doc_url

        return result
